#ifdef WIN32
#pragma warning(disable:4996)
#endif

#include <ctime>
#include <iostream>
#include <fstream>
#include <ctime>
#include <filesystem>
#include <regex>

#include "log.hpp"
#include "filesystem.hpp"

namespace fcf{
namespace details{


  log_t::log_t()
    : _stop(0), _level(LOG_LEVEL_LOG), _life_time(0) {
    _gc_thread = std::thread(&log_t::_gc, this);
  }

  log_t::~log_t() {
    {
      std::unique_lock<std::mutex> lock(_mutex);
      _stop = true;
      _condition.notify_all();
    }
    if (_gc_thread.joinable())
      _gc_thread.join();
  }

  void log_t::write(int a_level, const std::string& a_str){
    std::string msg;

    std::time_t time = std::time(0);
    std::tm* now = std::localtime(&time);
    char tmstr [80];
    std::strftime(tmstr, 80, "%F %T", now);
    msg += tmstr;
    msg += " ";

    switch(a_level){
      case LOG_LEVEL_TST: msg += "[TST]"; break;
      case LOG_LEVEL_ERR: msg += "[ERR]"; break;
      case LOG_LEVEL_WRN: msg += "[WRN]"; break;
      case LOG_LEVEL_LOG: msg += "[LOG]"; break;
      case LOG_LEVEL_TRC: msg += "[TRC]"; break;
    }

    msg += " [SERVER]:  ";
    msg += a_str;
  
    std::unique_lock<std::mutex> lock(_mutex);
    std::cout << msg << std::endl;

    if (!_file_prefix.empty()){
      std::strftime(tmstr, 80, "%F", now);
      std::string file_path(_file_prefix + tmstr + ".log");
      std::ofstream ofs(fcf::utf8str_to_path(file_path), std::ofstream::out | std::ofstream::app);
      ofs << msg << std::endl;
      ofs.close();
    }
  }

  void log_t::set_level(int a_level){
    std::unique_lock<std::mutex> lock(_mutex);
    _level = a_level;
  }

  int log_t::get_level(){
    return _level;
  }

  void log_t::set_file_prefix(const std::string& a_file_prefix){
    std::unique_lock<std::mutex> lock(_mutex);
    _file_prefix = a_file_prefix;
  }

  std::string log_t::get_file_preffix(){
    return _file_prefix;
  }

  void log_t::set_life_time(int a_life_time){
    std::unique_lock<std::mutex> lock(_mutex);
    _life_time = a_life_time;
  }

  int log_t::get_life_time(){
    return _life_time;
  }

  void log_t::remove_old_files(){
    std::unique_lock<std::mutex> lock(_mutex);
    _remove_old_files(lock);
  }

  void log_t::_remove_old_files(std::unique_lock<std::mutex>& a_lock){
    if (!_life_time)
      return;

    std::string           start_file_name;
    std::filesystem::path directory;
    std::filesystem::path prefix_path = fcf::utf8str_to_path(_file_prefix);
    if (!_file_prefix.empty() && _file_prefix[_file_prefix.length()-1] != '/' && _file_prefix[_file_prefix.length()-1] != '\\'){
      start_file_name = fcf::path_to_utf8str(prefix_path.filename());
      directory = prefix_path.parent_path();
    } else {
      directory = prefix_path;
    }

    for(auto& it: std::filesystem::directory_iterator(directory)) {
      std::string file_name = fcf::path_to_utf8str(it.path().filename());
      if (!start_file_name.empty() && file_name.find(start_file_name) ==  std::string::npos)
        continue;
      if (file_name.length() <= start_file_name.length())
        continue;

      // suffix
      // 2020-01-01.log
      std::string file_suffix = file_name.substr(start_file_name.length());
      std::regex  pieces_regex("([0-9][0-9][0-9][0-9])-([0-9][0-9])-([0-9][0-9]).log");
      std::smatch pieces_match;
      if (!std::regex_match(file_suffix, pieces_match, pieces_regex))
        continue;

      std::tm filetm;
      memset(&filetm, 0, sizeof(filetm));
      filetm.tm_year = std::stoi(pieces_match[1]) - 1900;
      filetm.tm_mon  = std::stoi(pieces_match[2]) - 1;
      filetm.tm_mday = std::stoi(pieces_match[3]);
      std::chrono::system_clock::time_point filetp = std::chrono::system_clock::from_time_t(std::mktime(&filetm));
      std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
      int days = std::chrono::duration_cast<std::chrono::hours>(now - filetp).count()/24;

      if (days >= _life_time) {
        a_lock.unlock();
        fcf::log << "Remove log file '" << fcf::path_to_utf8str(it.path()) << "'";
        a_lock.lock();
        std::error_code ec;
        std::filesystem::remove(it.path(), ec);
      }

    }
  }

  void log_t::_gc(){
    std::unique_lock<std::mutex> lock(_mutex);
    _condition.wait_for(lock, std::chrono::minutes(1));
    while(!_stop) {
      try {
        _remove_old_files(lock);
      } catch(...){
      }
      _condition.wait_for(lock, std::chrono::hours(1));
    }
  }

  log_t logger;

} // details namespace

details::log_input_t tst(details::logger, LOG_LEVEL_TST);
details::log_input_t err(details::logger, LOG_LEVEL_ERR);
details::log_input_t wrn(details::logger, LOG_LEVEL_WRN);
details::log_input_t log(details::logger, LOG_LEVEL_LOG);
details::log_input_t trc(details::logger, LOG_LEVEL_TRC);

} // fcf namespace